#include <stdint.h>
#include <string.h>
#include "fds.h"
#include "fds_util.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"
#include "nrf.h"
#include "nordic_common.h"

/* Flag to check fds initialization. */
static bool volatile m_fds_initialized;

/* Array to map FDS return values to strings. */
char const * fds_err_str[] =
{
    "FDS_SUCCESS",
    "FDS_ERR_OPERATION_TIMEOUT",
    "FDS_ERR_NOT_INITIALIZED",
    "FDS_ERR_UNALIGNED_ADDR",
    "FDS_ERR_INVALID_ARG",
    "FDS_ERR_NULL_ARG",
    "FDS_ERR_NO_OPEN_RECORDS",
    "FDS_ERR_NO_SPACE_IN_FLASH",
    "FDS_ERR_NO_SPACE_IN_QUEUES",
    "FDS_ERR_RECORD_TOO_LARGE",
    "FDS_ERR_NOT_FOUND",
    "FDS_ERR_NO_PAGES",
    "FDS_ERR_USER_LIMIT_REACHED",
    "FDS_ERR_CRC_CHECK_FAILED",
    "FDS_ERR_BUSY",
    "FDS_ERR_INTERNAL",
};

/* Array to map FDS events to strings. */
static char const * fds_evt_str[] =
{
    "FDS_EVT_INIT",
    "FDS_EVT_WRITE",
    "FDS_EVT_UPDATE",
    "FDS_EVT_DEL_RECORD",
    "FDS_EVT_DEL_FILE",
    "FDS_EVT_GC",
};

/* Keep track of the progress of a delete_all operation. */
static struct
{
    bool delete_next;   //!< Delete next record.
    bool pending;       //!< Waiting for an fds FDS_EVT_DEL_RECORD event, to delete the next record.
} m_delete_all;

/**@brief   Sleep until an event is received. */
static void sleep_during_setup_fds(void)
{
#ifdef SOFTDEVICE_PRESENT
    (void) sd_app_evt_wait();
#else
    __WFE();
#endif
}


/**@brief   Wait for fds to initialize. */
static void wait_for_fds_ready(void)
{
    while (!m_fds_initialized)
    {
        sleep_during_setup_fds();
    }
}

static void fds_evt_handler(fds_evt_t const * p_evt)
{
    NRF_LOG_INFO("Event: %s received (%s)",
                  fds_evt_str[p_evt->id],
                  fds_err_str[p_evt->result]);

    switch (p_evt->id)
    {
        case FDS_EVT_INIT:
            if (p_evt->result == FDS_SUCCESS)
            {
                m_fds_initialized = true;
            }
            break;

        case FDS_EVT_WRITE:
        {
            if (p_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_INFO("Record ID:\t0x%04x",  p_evt->write.record_id);
                NRF_LOG_INFO("File ID:\t0x%04x",    p_evt->write.file_id);
                NRF_LOG_INFO("Record key:\t0x%04x", p_evt->write.record_key);
            }
        } break;

        case FDS_EVT_DEL_RECORD:
        {
            if (p_evt->result == FDS_SUCCESS)
            {
                NRF_LOG_INFO("Record ID:\t0x%04x",  p_evt->del.record_id);
                NRF_LOG_INFO("File ID:\t0x%04x",    p_evt->del.file_id);
                NRF_LOG_INFO("Record key:\t0x%04x", p_evt->del.record_key);
            }
            m_delete_all.pending = false;
        } break;

        default:
            break;
    }
}


/**@brief   Begin deleting all records, one by one. */
void delete_all_begin(void)
{
    m_delete_all.delete_next = true;
}

bool record_delete_next(void)
{
    fds_find_token_t  tok   = {0};
    fds_record_desc_t desc  = {0};

    if (fds_record_iterate(&desc, &tok) == FDS_SUCCESS)
    {
        ret_code_t rc = fds_record_delete(&desc);
        if (rc != FDS_SUCCESS)
        {
            return false;
        }

        return true;
    }
    else
    {
        /* No records left to delete. */
        return false;
    }
}

/**@brief   Process a delete all command.
 *
 * Delete records, one by one, until no records are left.
 */
void delete_all_process(void)
{
    if (   m_delete_all.delete_next
        & !m_delete_all.pending)
    {
        NRF_LOG_INFO("Deleting next record.");
        NRF_LOG_FLUSH();

        m_delete_all.delete_next = record_delete_next();
        if (!m_delete_all.delete_next)
        {
            NRF_LOG_DEBUG("No records left to delete.");
            NRF_LOG_FLUSH();

        }
    }
}

uint32_t fds_read(uint16_t config_file, uint16_t config_key, uint8_t *o_data, uint32_t o_len)
{
  ret_code_t rc;

  fds_record_desc_t desc = {0};
  fds_find_token_t  tok  = {0};

  memset(&tok, 0x00, sizeof(fds_find_token_t));

  rc = fds_record_find(config_file, config_key, &desc, &tok);

  if (rc == FDS_SUCCESS)
  {
      /* A config file is in flash. Let's update it. */
      fds_flash_record_t config = {0};

      /* Open the record and read its contents. */
      rc = fds_record_open(&desc, &config);
      APP_ERROR_CHECK(rc);

      NRF_LOG_INFO("Key Found, reading");
      NRF_LOG_FLUSH();
      /* Copy the configuration from flash */
      memcpy(o_data, config.p_data, sizeof(uint8_t) * o_len);

      /* Close the record when done reading. */
      rc = fds_record_close(&desc);
      APP_ERROR_CHECK(rc);

      //NRF_LOG_INFO("output = %s\r\n", o_data);
      //NRF_LOG_FLUSH();
      return 0;
  }

  return 1;
}

void fds_write(uint16_t config_file, uint16_t config_key, uint8_t *i_data, uint32_t i_len)
{
  ret_code_t rc;

  /* A record containing dummy configuration data. */
  fds_record_t fds_record =
  {
      .file_id           = config_file,
      .key               = config_key,
      .data.p_data       = i_data,
      /* The length of a record is always expressed in 4-byte units (words). */
      .data.length_words = (sizeof(uint8_t) * i_len + 3) / sizeof(uint32_t),
  };

  fds_record_desc_t desc = {0};
  fds_find_token_t  tok  = {0};

  rc = fds_record_find(config_file, config_key, &desc, &tok);

  if (rc == FDS_SUCCESS)
  {
      NRF_LOG_INFO("File found, updating record to %s.", i_data);
      NRF_LOG_FLUSH();

      /* Write the updated record to flash. */
      rc = fds_record_update(&desc, &fds_record);
      APP_ERROR_CHECK(rc);
  }
  else
  {
      /* System config not found; write a new one. */
      NRF_LOG_INFO("Writing config file...");
      NRF_LOG_FLUSH();

      rc = fds_record_write(&desc, &fds_record);
      APP_ERROR_CHECK(rc);
  }
}

void fds_check_stat(void)
{
  ret_code_t rc;
  fds_stat_t stat = {0};

  rc = fds_stat(&stat);
  APP_ERROR_CHECK(rc);

  NRF_LOG_INFO("Found %d valid records.", stat.valid_records);
  NRF_LOG_FLUSH();

  NRF_LOG_INFO("Found %d dirty records (ready to be garbage collected).", stat.dirty_records);
  NRF_LOG_FLUSH();

}

void fds_garbage_collect(void)
{
  ret_code_t rc;

  rc = fds_gc();
  APP_ERROR_CHECK(rc);
}

void fds_setup(void)
{

  ret_code_t rc;
  /* Register first to receive an event when initialization is complete. */
  (void) fds_register(fds_evt_handler);

  NRF_LOG_INFO("Initializing fds...");
  NRF_LOG_FLUSH();

  rc = fds_init();
  APP_ERROR_CHECK(rc);

  wait_for_fds_ready();

}
